/* Atomixed -- the level-/themeeditor for atomix.
 * Copyright (C) 1999-2000 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "image_win_ctrl.h"
#include "canvas_helper.h"
#include "support.h"
#include "tile.h"

/*=================================================================
 
  Global variables
  
  ---------------------------------------------------------------*/
static GtkWidget *window;
static Theme *actual_theme;
static Tile *actual_tile;

typedef struct {
	GnomeCanvasGroup *moveables;
	GnomeCanvasGroup *connections;
	GnomeCanvasGroup *obstacles;

	GHashTable *moveable_items;
	GHashTable *connection_items;
	GHashTable *obstacle_items;
	
	GnomeCanvasItem *selected_moveable;
	GnomeCanvasItem *selected_obstacle;	
} PreviewData;

static PreviewData *preview_data;

typedef enum {
	PREVIEW_OBSTACLE,
	PREVIEW_MOVEABLE,
	PREVIEW_CONNECTION
} PreviewType;

typedef	struct {
	GnomeIconList *gil;
	PreviewType type;
} AddImageData;


/*=================================================================
 
  Declaration of internal functions

  ---------------------------------------------------------------*/

void image_win_ctrl_clear_preview_data();

void image_win_ctrl_add_preview_item(GdkImlibImage *image, gint id, PreviewType ptype);

void image_win_ctrl_render_preview(Tile *tile);

void image_win_ctrl_add_image_to_clist(gpointer key, gpointer value, gpointer user_data);

void hide_canvas_item(gpointer key, gpointer value, gpointer user_data);
/*=================================================================
 
  Image Window Ctrl initialisation and clean up

  ---------------------------------------------------------------*/

void image_win_ctrl_init(GtkWidget *win)
{
	const gint row_spacing = 20;
	const gint col_spacing = 20;
	const gint text_spacing = 5;
	GnomeIconList *gil_obstacle;
	GnomeIconList *gil_moveable;
	GnomeIconList *gil_connection;

	window = win;

	/* init obastcle section */
	gil_obstacle = GNOME_ICON_LIST(lookup_widget(window, "obstacle_list"));
	gnome_icon_list_set_row_spacing (gil_obstacle,row_spacing);
        gnome_icon_list_set_col_spacing (gil_obstacle, col_spacing);
        gnome_icon_list_set_text_spacing (gil_obstacle, text_spacing);

	/* init moveable section */
	gil_moveable = GNOME_ICON_LIST(lookup_widget(window, "moveable_list"));
	gnome_icon_list_set_row_spacing (gil_moveable, row_spacing);
        gnome_icon_list_set_col_spacing (gil_moveable, col_spacing);
        gnome_icon_list_set_text_spacing (gil_moveable, text_spacing);

	/* fill connection section */	
	gil_connection = GNOME_ICON_LIST(lookup_widget(window, "connection_list"));
	gnome_icon_list_set_row_spacing (gil_connection, row_spacing);
        gnome_icon_list_set_col_spacing (gil_connection, col_spacing);
        gnome_icon_list_set_text_spacing (gil_connection, text_spacing);
	gnome_icon_list_set_selection_mode(gil_connection, GTK_SELECTION_MULTIPLE);

	actual_tile = tile_new();
}

void image_win_ctrl_destroy()
{
	tile_destroy(actual_tile);
	if(preview_data)
	{
		g_hash_table_destroy(preview_data->moveable_items);
		g_hash_table_destroy(preview_data->connection_items);
		g_hash_table_destroy(preview_data->obstacle_items);
		g_free(preview_data);
	}
}

/*=================================================================
  
  Image Window Ctrl functions

  ---------------------------------------------------------------*/

void image_win_ctrl_set_theme(Theme *theme)
{
	GnomeIconList *gil_obstacle;
	GnomeIconList *gil_moveable;
	GnomeIconList *gil_connection;
	GnomeCanvas *preview_canvas;
	gint tile_width, tile_height;	
	AddImageData *add_image_data;
	
	add_image_data = g_malloc(sizeof(AddImageData));
	theme_get_tile_size(theme, &tile_width, &tile_height);
	actual_theme = theme;
	image_win_ctrl_clear_preview_data();	

	/* fill obstacle section */
	gil_obstacle = GNOME_ICON_LIST(lookup_widget(window, "obstacle_list"));
	gnome_icon_list_freeze(gil_obstacle);
	gnome_icon_list_set_icon_width(gil_obstacle, tile_width); 
	gnome_icon_list_clear(gil_obstacle);
	gnome_icon_list_unselect_all(gil_obstacle, NULL, NULL);
	add_image_data->gil = gil_obstacle;
	add_image_data->type = PREVIEW_OBSTACLE;
	g_hash_table_foreach(theme->obstacles, 
			     (GHFunc) image_win_ctrl_add_image_to_clist, 
			     (gpointer) add_image_data);
	gnome_icon_list_thaw(gil_obstacle);

	/* fill moveable section */
	gil_moveable = GNOME_ICON_LIST(lookup_widget(window, "moveable_list"));
	gnome_icon_list_freeze(gil_moveable);
	gnome_icon_list_set_icon_width(gil_moveable, tile_width); 
	gnome_icon_list_clear(gil_moveable);
	gnome_icon_list_unselect_all(gil_moveable, NULL, NULL);
	add_image_data->gil = gil_moveable;
	add_image_data->type = PREVIEW_MOVEABLE;
	g_hash_table_foreach(theme->moveables, 
			     (GHFunc) image_win_ctrl_add_image_to_clist, 
			     (gpointer) add_image_data);
	gnome_icon_list_thaw(gil_moveable);

	/* fill connection section */	
	gil_connection = GNOME_ICON_LIST(lookup_widget(window, "connection_list"));
	gnome_icon_list_freeze(gil_connection);
	gnome_icon_list_clear(gil_connection);
	gnome_icon_list_set_icon_width(gil_connection, tile_width); 
	gnome_icon_list_unselect_all(gil_connection, NULL, NULL);
	add_image_data->gil = gil_connection;
	add_image_data->type = PREVIEW_CONNECTION;
	g_hash_table_foreach(theme->connections, 
			     (GHFunc) image_win_ctrl_add_image_to_clist, 
			     (gpointer) add_image_data);
	gnome_icon_list_thaw(gil_connection);
	
	/* try to set background color */
	/* FIXME: doesn't work! */
	set_background_color_ref(GTK_WIDGET(gil_obstacle), 
				 theme_get_background_color(theme));
	set_background_color_ref(GTK_WIDGET(gil_moveable), 
				 theme_get_background_color(theme));

	/* init preview canvas*/
	preview_canvas = GNOME_CANVAS(lookup_widget(window, "atom_preview"));
	set_background_color_ref(GTK_WIDGET(preview_canvas), 
				 theme_get_background_color(theme));	
	gnome_canvas_set_scroll_region(preview_canvas, 0.0, 0.0, tile_width, tile_height);
}

void image_win_ctrl_select_obstacle(gint icon)
{
	GnomeIconList *gil_obstacle;
	gint image_id;

	gil_obstacle = GNOME_ICON_LIST(lookup_widget(window, "obstacle_list"));	

	if(tile_get_type(actual_tile) != TILE_OBSTACLE)
	{
		GnomeIconList *gil_moveable;
		GnomeIconList *gil_connection;
		gil_moveable = GNOME_ICON_LIST(lookup_widget(window, "moveable_list"));
		gil_connection = GNOME_ICON_LIST(lookup_widget(window, "connection_list"));

		gnome_icon_list_unselect_all(gil_moveable, NULL, NULL);
		gnome_icon_list_unselect_all(gil_connection, NULL, NULL);

		tile_set_type(actual_tile, TILE_OBSTACLE);
	}

	image_id = GPOINTER_TO_INT(gnome_icon_list_get_icon_data(gil_obstacle,icon));
	tile_set_image_id(actual_tile, image_id);	
	image_win_ctrl_render_preview(actual_tile);
}


void image_win_ctrl_select_moveable(gint icon)
{
	GnomeIconList *gil_moveable;
	gint image_id;

	gil_moveable = GNOME_ICON_LIST(lookup_widget(window, "moveable_list"));	
	g_return_if_fail(gil_moveable != NULL);

	if(tile_get_type(actual_tile) != TILE_MOVEABLE)
	{
		GnomeIconList *gil_obstacle;
		gil_obstacle = GNOME_ICON_LIST(lookup_widget(window, "obstacle_list"));
		gnome_icon_list_unselect_all(gil_obstacle, NULL, NULL);

		tile_set_type(actual_tile, TILE_MOVEABLE);
	}

	image_id = GPOINTER_TO_INT(gnome_icon_list_get_icon_data(gil_moveable,icon));
	tile_set_image_id(actual_tile, image_id);	
	image_win_ctrl_render_preview(actual_tile);	
}

void image_win_ctrl_select_connection(gint icon)
{
	GnomeIconList *gil_connection;
	gint image_id;

	gil_connection = GNOME_ICON_LIST(lookup_widget(window, "connection_list"));	
	g_return_if_fail(gil_connection!= NULL);

	/* add image id to tile */
	image_id = GPOINTER_TO_INT(gnome_icon_list_get_icon_data(gil_connection,
								 icon));
	tile_add_connection_id(actual_tile, image_id);
	image_win_ctrl_render_preview(actual_tile);
}

void image_win_ctrl_unselect_connection(gint icon)
{
	GnomeIconList *gil_connection;
	gint image_id;
	
	gil_connection = GNOME_ICON_LIST(lookup_widget(window, "connection_list"));	
	g_return_if_fail(gil_connection!= NULL);

	/* remove image id from tile list*/
	image_id = GPOINTER_TO_INT(gnome_icon_list_get_icon_data(gil_connection,
								 icon));
	tile_remove_connection_id(actual_tile, image_id);
	image_win_ctrl_render_preview(actual_tile);
}

Tile* image_win_ctrl_get_actual_tile()
{
	return actual_tile;
}


/*=================================================================
 
  Internal helper functions

  ---------------------------------------------------------------*/

void
image_win_ctrl_add_image_to_clist(gpointer key, gpointer value, gpointer user_data)
{
	ThemeElement *element = (ThemeElement*) value;
	AddImageData *data = (AddImageData*) user_data;
	GnomeIconList *gil = data->gil;
	gint id = GPOINTER_TO_INT(key);
	gint pos; 
	
	GdkImlibImage *image = theme_get_element_image(actual_theme, element);

 	pos = gnome_icon_list_append_imlib(gil, image, "test");
	gnome_icon_list_set_icon_data(gil, pos, GINT_TO_POINTER(id));
	
	image_win_ctrl_add_preview_item(image, id, data->type);
}

void image_win_ctrl_clear_preview_data()
{
	GnomeCanvas *canvas;

	canvas = GNOME_CANVAS(lookup_widget(window, "atom_preview"));

	if(preview_data)
	{
		if(preview_data->moveables)
		{
			gtk_object_destroy(GTK_OBJECT(preview_data->moveables));
		}
		if(preview_data->connections)
		{
			gtk_object_destroy(GTK_OBJECT(preview_data->connections));
		}
		if(preview_data->obstacles)
		{
			gtk_object_destroy(GTK_OBJECT(preview_data->obstacles));
		}		
		if(preview_data->moveable_items)
		{
			g_hash_table_destroy(preview_data->moveable_items);
		}
		if(preview_data->connection_items)
		{
			g_hash_table_destroy(preview_data->connection_items);
		}
		if(preview_data->obstacle_items)
		{
			g_hash_table_destroy(preview_data->obstacle_items);
		}
	}
	else
	{
		preview_data = g_malloc(sizeof(PreviewData));
	}
	
	preview_data->moveables = create_group_ref(canvas, NULL);
	gnome_canvas_item_hide(GNOME_CANVAS_ITEM(preview_data->moveables));
	preview_data->connections = create_group_ref(canvas, NULL);
	gnome_canvas_item_hide(GNOME_CANVAS_ITEM(preview_data->connections));
	preview_data->obstacles = create_group_ref(canvas, NULL);
	gnome_canvas_item_hide(GNOME_CANVAS_ITEM(preview_data->obstacles));

	gnome_canvas_item_lower_to_bottom(GNOME_CANVAS_ITEM(preview_data->connections));
	gnome_canvas_item_raise_to_top(GNOME_CANVAS_ITEM(preview_data->obstacles));
	
	preview_data->moveable_items = g_hash_table_new((GHashFunc) g_direct_hash, 
							(GCompareFunc) g_direct_equal);
	preview_data->connection_items =  g_hash_table_new((GHashFunc) g_direct_hash, 
							   (GCompareFunc) g_direct_equal);
	preview_data->obstacle_items =  g_hash_table_new((GHashFunc) g_direct_hash, 
							 (GCompareFunc) g_direct_equal);
	
	preview_data->selected_moveable = NULL;
	preview_data->selected_obstacle = NULL;
}

void image_win_ctrl_add_preview_item(GdkImlibImage *image, gint id, PreviewType ptype)
{
	GnomeCanvasGroup *group = NULL;
	GnomeCanvasItem *item;
	
	switch(ptype)
	{
	case PREVIEW_OBSTACLE:
		group = preview_data->obstacles;
		break;

	case PREVIEW_MOVEABLE:
		group = preview_data->moveables;
		break;

	case PREVIEW_CONNECTION:
		group = preview_data->connections;
		break;
		
	default:
		g_print("image_win_ctrl.c: We shouldn`t be here!\n");
	}

 	item = gnome_canvas_item_new(group,
				     gnome_canvas_image_get_type(),
				     "image", image,
				     "x", 0.0,
				     "y", 0.0,
				     "width", (double)image->rgb_width,
				     "height", (double)image->rgb_height,
				     "anchor", GTK_ANCHOR_NW,
				     NULL);                              
	gnome_canvas_item_hide(item);

	switch(ptype)
	{
	case PREVIEW_OBSTACLE:
	       g_hash_table_insert(preview_data->obstacle_items,
				   GINT_TO_POINTER(id),
				   (gpointer)item);
		break;

	case PREVIEW_MOVEABLE:
		g_hash_table_insert(preview_data->moveable_items,
				    GINT_TO_POINTER(id),
				    (gpointer)item);
		break;

	case PREVIEW_CONNECTION:
	        g_hash_table_insert(preview_data->connection_items,
				    GINT_TO_POINTER(id),
				    (gpointer)item);
		break;
		
	default:
		g_print("image_win_ctrl.c: We shouldn`t be here!\n");
	}
}

void image_win_ctrl_render_preview(Tile *tile)
{
	if(preview_data)
	{
		GnomeCanvasItem *item;
		TileType type = tile_get_type(tile);

		if(type == TILE_MOVEABLE)
		{
			GSList *tile_conn_ids = NULL;

			/* hide all obstacles */
			gnome_canvas_item_hide(GNOME_CANVAS_ITEM(preview_data->obstacles));

			/* show atom image */
			if(preview_data->selected_moveable)
			{
				gnome_canvas_item_hide(preview_data->selected_moveable);
			}
			
			item = GNOME_CANVAS_ITEM(
				g_hash_table_lookup(preview_data->moveable_items,
						    GINT_TO_POINTER(tile_get_image_id(tile))));
			gnome_canvas_item_show(item);
			preview_data->selected_moveable = item;
			gnome_canvas_item_show(GNOME_CANVAS_ITEM(preview_data->moveables));

			/* first hide all connections ... */
			g_hash_table_foreach(preview_data->connection_items,
					     (GHFunc) hide_canvas_item, NULL);
			
			/* ... then show the selected connection images */
			tile_conn_ids = tile_get_connection_ids(tile);
			while(tile_conn_ids != NULL)
			{
				gint id = GPOINTER_TO_INT(tile_conn_ids->data);
				item = GNOME_CANVAS_ITEM(
					g_hash_table_lookup(preview_data->connection_items,
							 GINT_TO_POINTER(id)));
				gnome_canvas_item_show(item);

				tile_conn_ids = tile_conn_ids->next;
			}
			gnome_canvas_item_show(GNOME_CANVAS_ITEM(preview_data->connections));
		}
		else if(type == TILE_OBSTACLE)
		{
			/* hide all moveables and connections */
			gnome_canvas_item_hide(GNOME_CANVAS_ITEM(preview_data->moveables));
			gnome_canvas_item_hide(GNOME_CANVAS_ITEM(preview_data->connections));

			/* hide previous shown obstacle */
			if(preview_data->selected_obstacle)
			{
				gnome_canvas_item_hide(preview_data->selected_obstacle);
			}
			
			/* show actual selected obstacle image*/
			item = GNOME_CANVAS_ITEM(
				g_hash_table_lookup(preview_data->obstacle_items, 
						    GINT_TO_POINTER(tile_get_image_id(tile))));
			gnome_canvas_item_show(item);
			preview_data->selected_obstacle = item;
			gnome_canvas_item_show(GNOME_CANVAS_ITEM(preview_data->obstacles));
		}
	}
}

void
hide_canvas_item(gpointer key, gpointer value, gpointer user_data)
{
	gnome_canvas_item_hide(GNOME_CANVAS_ITEM(value));
}
