/* Atomixed -- the level-/themeeditor for atomix.
 * Copyright (C) 1999-2000 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "playfield_editor.h"
#include "canvas_helper.h"
#include "image_win_ctrl.h"
#include "level_ctrl.h"
#include "support.h"
#include "tile.h"
#include "main.h"
#include <gnome.h>
#include <math.h>


/*=================================================================
 
  Declaration of internal functions

  ---------------------------------------------------------------*/
void playfield_editor_render(PlayFieldEditor *pfe);

void playfield_editor_create_grid(PlayFieldEditor *pfe);

void playfield_editor_clear_grid(PlayFieldEditor *pfe);

GnomeCanvasItem* create_obstacle (PlayFieldEditor *pfe, 
				  double x, double y, Tile *tile);

GnomeCanvasItem* create_moveable (PlayFieldEditor *pfe, 
				  double x, double y, Tile *tile);

GnomeCanvasItem* create_item (GdkImlibImage* img, double x, double y,
			      GnomeCanvasGroup* group, PlayFieldEditor *pfe);
gint item_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data);

gint grid_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data);

void set_tile(GnomeCanvasItem *grid_item, PlayFieldEditor *pfe);

void remove_tile(GnomeCanvasItem *item, PlayFieldEditor *pfe);

void replace_tile(GnomeCanvasItem *item, PlayFieldEditor *pfe);

/*=================================================================
 
  Creation, initialisation and clean up

  ---------------------------------------------------------------*/

PlayFieldEditor* playfield_editor_new(gchar *canvas_name, 
				      gboolean anm)
{
	PlayFieldEditor *pfe = NULL;

	pfe = g_malloc(sizeof(PlayFieldEditor));
	
	if(pfe)
	{
		GnomeCanvas *canvas;

		pfe->canvas_name = g_strdup(canvas_name);
		pfe->theme = NULL;
		pfe->allow_non_moveables = anm;
		pfe->show_grid = TRUE;
		pfe->pf = NULL;
		
		canvas = GNOME_CANVAS(lookup_widget(get_app(), canvas_name));
		pfe->grid = create_group_ref(canvas, NULL);
		pfe->obstacles = create_group_ref(canvas, NULL);
		pfe->moveables = create_group_ref(canvas, NULL);
		gnome_canvas_item_lower_to_bottom(GNOME_CANVAS_ITEM(pfe->grid));

	}
	
	return pfe;
}

void playfield_editor_init(PlayFieldEditor *pfe)
{
}


void playfield_editor_destroy(PlayFieldEditor *pfe)
{
	if(pfe)
	{
		if(pfe->moveables)
		{
			gtk_object_destroy(GTK_OBJECT(pfe->moveables));
		}
		if(pfe->obstacles)
		{
			gtk_object_destroy(GTK_OBJECT(pfe->obstacles));
		}
		g_free(pfe);
	}
}

/*=================================================================
  
  Editor functions

  ---------------------------------------------------------------*/

void playfield_editor_clear(PlayFieldEditor *pfe)
{
	if(pfe->moveables)
	{
		gtk_object_destroy(GTK_OBJECT(pfe->moveables));
	}
	if(pfe->obstacles)
	{
		gtk_object_destroy(GTK_OBJECT(pfe->obstacles));
	}
	pfe->obstacles = create_group(pfe->canvas_name, NULL);
	pfe->moveables = create_group(pfe->canvas_name, NULL);	
}

void playfield_editor_set_playfield(PlayFieldEditor *pfe, PlayField *pf)
{
	if(pfe)
	{
		pfe->pf = pf;		
		playfield_editor_clear(pfe);
		playfield_editor_clear_grid(pfe);
		playfield_editor_render(pfe);
		playfield_editor_create_grid(pfe);	       		
	}
	else
	{
		g_print("playfield_editor_set: pfe == NULL");
	}
}

void playfield_editor_resize(PlayFieldEditor *pfe, guint n_rows, guint n_cols)
{
	g_return_if_fail(pfe!=NULL);

	if((pfe->pf->n_rows != n_rows) || (pfe->pf->n_cols != n_cols))
	{
		g_print("Resize Playfield to %ix%i\n", n_rows, n_cols);	
		playfield_set_matrix_size(pfe->pf, n_rows, n_cols);
		playfield_editor_set_playfield(pfe, pfe->pf);

		/* mark level as modified */
		level_ctrl_set_actual_level_modified();
	}
}


void playfield_editor_set_theme(PlayFieldEditor *pfe, Theme *theme)
{
	if(pfe && theme)
	{
		pfe->theme = theme;
	}
}


void playfield_editor_show_grid(PlayFieldEditor *pfe, gboolean show)
{
	if(show)
	{
		gnome_canvas_item_show(GNOME_CANVAS_ITEM(pfe->grid));		
	}
	else
	{
		gnome_canvas_item_hide(GNOME_CANVAS_ITEM(pfe->grid));
	}
	pfe->show_grid = show;
}


/*=================================================================
 
  Internal creation functions

  ---------------------------------------------------------------*/
GnomeCanvasItem*
create_obstacle (PlayFieldEditor *pfe, double x, double y, Tile *tile)
{
	GdkImlibImage *img = NULL;
	GnomeCanvasItem *item = NULL;
	
	img = theme_get_tile_image(pfe->theme, tile);
	
	if(img) {
		item = create_item (img, x, y, pfe->obstacles, pfe);
		gtk_signal_connect(GTK_OBJECT(item), "event",
				   (GtkSignalFunc) item_event, pfe);
	} 
#ifdef DEBUG
	else {
		g_print("Obstacle image not found!\n");
	}
#endif    
	
	return item;
}

GnomeCanvasItem*
create_moveable (PlayFieldEditor *pfe, double x, double y, Tile *tile)
{
	GdkImlibImage *img = NULL;
	GdkImlibImage  *conn_img = NULL;
	GnomeCanvasItem *item = NULL;
	GnomeCanvasGroup *moveable_group = NULL;
	GSList *conn_images;
	
	img = theme_get_tile_image(pfe->theme, tile);
	conn_images = theme_get_tile_connection_images(pfe->theme, tile);
		
	if(img) 
	{
		gint i;

		/* create new group at (0,0) */
		moveable_group = create_group(pfe->canvas_name, pfe->moveables);

		/* add connection and moveable images */
		if(conn_images)
		{
			for(i = 0; i < g_slist_length(conn_images); i++)
			{
				conn_img = (GdkImlibImage*) g_slist_nth_data(conn_images, i);
				item = create_item(conn_img, 0.0, 0.0, moveable_group,
					pfe);
			}
		}
		item = create_item (img, 0.0, 0.0, moveable_group, pfe);
		gtk_signal_connect(GTK_OBJECT(moveable_group), "event",
				   (GtkSignalFunc) item_event, pfe);

		/* move to the right location */
		gnome_canvas_item_move(GNOME_CANVAS_ITEM(moveable_group), x, y);
	} 
#ifdef DEBUG
	else {
		g_print("Atom image not found!\n");
	}
#endif    
	g_slist_free(conn_images);

	return GNOME_CANVAS_ITEM(moveable_group);
}

GnomeCanvasItem*
create_item (GdkImlibImage* img, double x, double y,
	     GnomeCanvasGroup* group, PlayFieldEditor *pfe)
{
	GnomeCanvasItem *item;
	
	item = gnome_canvas_item_new(group,
				     gnome_canvas_image_get_type(),
				     "image", img,
				     "x", x,
				     "y", y,
				     "width", (double)img->rgb_width,
				     "height", (double)img->rgb_height,
				     "anchor", GTK_ANCHOR_NW,
				     NULL);                              
	return item;
}

/*=================================================================
 
  Mouse handling functions

  ---------------------------------------------------------------*/

gint item_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
{
	PlayFieldEditor *pfe = (PlayFieldEditor*) data;

	switch (event->type) {
	case GDK_BUTTON_PRESS:		
	    if (event->button.button == 1) 
	    {
		    replace_tile(item, pfe);
	    }
	    else if (event->button.button == 3) 
	    {
		    g_print("right mouse button pressed on item.\n");
		    remove_tile(item, pfe);
	    }
	    break;

	default:
	}

	return TRUE;
}

gint grid_event (GnomeCanvasItem *item, GdkEvent *event, gpointer data)
{
	PlayFieldEditor *pfe = (PlayFieldEditor*) data;
	
	switch (event->type) {
	case GDK_BUTTON_PRESS:		
	    if (event->button.button == 1) 
	    {
		    set_tile(item, pfe);
	    }
	    else if (event->button.button == 3) 
	    {
		    g_print("right mouse button pressed on grid.\n");
	    }
	    break;

	default:
	}

	return TRUE;
}

/*=================================================================
 
  Functions for setting/removing tiles.

  ---------------------------------------------------------------*/
void replace_tile(GnomeCanvasItem *item,   PlayFieldEditor *pfe)
{
	gdouble x1, y1, x2, y2;	
	gint row, col;
	Tile *tile;
	Tile *tile_cp;
	GnomeCanvasItem *new_item = NULL;

	/* get position */
	gnome_canvas_item_get_bounds(item, &x1, &y1, &x2, &y2);       
	convert_to_playfield(pfe->theme, x1, y1, &row, &col);

	/* get tile and check whether its allowed to be set */
	tile = image_win_ctrl_get_actual_tile();
	if((!pfe->allow_non_moveables) && (tile_get_type(tile) == TILE_OBSTACLE))
	{
		return;
	}

	/* put the current selected item into the playfield structure */
	tile_cp = tile_copy(tile);
	playfield_set_tile(pfe->pf, row, col, tile_cp);

	/* destroy old canvas item */
	gtk_object_destroy(GTK_OBJECT(item));

	/* put it in the canvas */
	convert_to_canvas(pfe->theme, row, col, &x1, &y1);
	switch(tile_get_type(tile))
	{
	case TILE_MOVEABLE:		
		new_item = create_moveable(pfe, x1, y1, tile);
		break;
	case TILE_OBSTACLE:
		new_item = create_obstacle(pfe, x1, y1, tile);
		break;
	default:
	}	

	/* mark level as modified */
	level_ctrl_set_actual_level_modified();
}

void set_tile(GnomeCanvasItem *grid_item,  PlayFieldEditor *pfe)
{
	gdouble x1, y1, x2, y2;	
	gint row, col;
	Tile *tile;
	Tile *tile_cp;
	Tile *tile_pf;
	GnomeCanvasItem *item = NULL;

	/* get position */
	gnome_canvas_item_get_bounds(grid_item, &x1, &y1, &x2, &y2);       
	convert_to_playfield(pfe->theme, x1, y1, &row, &col);

	/* get tile and check whether its allowed to be set */
	tile = image_win_ctrl_get_actual_tile();
	if((!pfe->allow_non_moveables) && (tile_get_type(tile) == TILE_OBSTACLE))
	{
		return;
	}

	/* check whether there is already an tile at this position. 
	 * The user can click on a grid item, even if the tile is 
	 * not really empty. */
	tile_pf = playfield_get_tile(pfe->pf, row, col);
	if(tile_get_type(tile_pf) != TILE_NONE)
	{
		return;
	}
	
	/* put the current assembled tile in the playfield structure */
	tile_cp = tile_copy(tile);
	playfield_set_tile(pfe->pf, row, col, tile_cp);

	/* put it in the canvas */
	convert_to_canvas(pfe->theme, row, col, &x1, &y1);
	switch(tile_get_type(tile))
	{
	case TILE_MOVEABLE:		
		item = create_moveable(pfe, x1, y1, tile);
		break;
	case TILE_OBSTACLE:
		item = create_obstacle(pfe, x1, y1, tile);
		break;
	default:
	}

	/* mark level as modified */
	level_ctrl_set_actual_level_modified();
}

void remove_tile(GnomeCanvasItem *item, PlayFieldEditor *pfe)
{
	gdouble x1, y1, x2, y2;
	gint row, col;
	Tile *tile;
	
	/* get position */
	gnome_canvas_item_get_bounds(item, &x1, &y1, &x2, &y2);
	convert_to_playfield(pfe->theme, x1, y1, 
			     &row, &col);
	
	/* remove from playfield */
	tile = playfield_clear_tile(pfe->pf, row, col);
	tile_destroy(tile);
	
	/* remove from view */
	gtk_object_destroy(GTK_OBJECT(item));

	/* mark level as modified */
	level_ctrl_set_actual_level_modified();
}


/*=================================================================
 
  Internal helper functions

  ---------------------------------------------------------------*/
void playfield_editor_render(PlayFieldEditor *pfe)
{
	if(pfe && pfe->pf)
	{
		PlayField *pf = pfe->pf;
		gint row, col;
		gint width, height;
		gint tile_width, tile_height;

		for(row = 0; row < pf->n_rows; row++)
		{
			for(col = 0; col < pf->n_cols; col++)
			{
				GnomeCanvasItem *item = NULL;
				Tile *tile = playfield_get_tile(pf, row, col);
				TileType type = tile_get_type(tile);
				gdouble x,y;

				switch(type)
				{
				case TILE_MOVEABLE:
					convert_to_canvas(pfe->theme, 
							  row, col, &x, &y);
					item = create_moveable(pfe, x, y, tile);
					break;
					
				case TILE_OBSTACLE:
					if(pfe->allow_non_moveables)
					{
						convert_to_canvas(pfe->theme, 
								  row, 
								  col, 
								  &x, &y);
						item = create_obstacle(pfe, 
								       x, y, 
								       tile);
					}
					break;
					
				case TILE_NONE:
				default:
				}				

			}
		}		
		theme_get_tile_size(pfe->theme, &tile_width, &tile_height);
		width = pf->n_cols * tile_width;
		height = pf->n_rows * tile_height;		
		set_canvas_dimensions(pfe->canvas_name, width, height);
		set_background_color(pfe->canvas_name, 
				     theme_get_background_color(pfe->theme));
	}
	else
	{
		g_print("playfield_editor_render: pfe or pfe->pf == NULL");
	}	
}

void playfield_editor_clear_grid(PlayFieldEditor *pfe)
{
	if(pfe->grid)
	{
		gtk_object_destroy(GTK_OBJECT(pfe->grid));
	}
	
	pfe->grid = create_group(pfe->canvas_name, NULL);
	gnome_canvas_item_lower_to_bottom(GNOME_CANVAS_ITEM(pfe->grid));
	if(!pfe->show_grid)
	{
		gnome_canvas_item_hide(GNOME_CANVAS_ITEM(pfe->grid));
	}
}

void playfield_editor_create_grid(PlayFieldEditor *pfe)
{
	gint tile_width, tile_height;
	gint row, col;
	gint max_row, max_col;
	gchar *bg_color;

	/* initialize grid */
	theme_get_tile_size(pfe->theme, &tile_width, &tile_height);
	bg_color = theme_get_background_color(pfe->theme);
	max_row = pfe->pf->n_rows;
	max_col = pfe->pf->n_cols;
	for(row = 0; row < max_row; row++)
	{
		for(col = 0; col < max_col; col++)
		{

			GnomeCanvasItem *item;
			gdouble x1, y1, x2, y2;

			x1 = col * tile_width;
			y1 = row * tile_height;
			x2 = x1 + tile_width;
			y2 = y1 + tile_height;

			item = gnome_canvas_item_new(pfe->grid,
						     gnome_canvas_rect_get_type(),
						     "x1", x1,
						     "y1", y1,
						     "x2", x2,
						     "y2", y2,
						     "fill_color", bg_color,
						     "outline_color", "white",
						     "width_pixels", 1,
						     NULL);
			gtk_signal_connect(GTK_OBJECT(item), "event",
					   (GtkSignalFunc) grid_event, pfe);
		}
	}
}

